//---------------------------------------------------------------------------
// MerchantableStemWoodVolume
//---------------------------------------------------------------------------
#if !defined(MerchantableStemWoodVolume_H)
  #define MerchantableStemWoodVolume_H

#include "BehaviorBase.h"

/**
* Merchantable stem wood volume calculator
*
* This class calculates merchantable wood volume from the FIADB manual.  This
* uses as an input biomass as calculated by the class clDimensionAnalysis.
*
* volume (ft3) of merchantable stem wood (not bark) = merchantable stem wood
* biomass (pounds) / WOOD_SPGR_GREENVOL_DRYWT, where WOOD_SPGR_GREENVOL_DRYWT
* is a parameter.
*
* Merchantable stem wood biomass = Biomass stem fraction * biomass * conversion
* factor from Mg to pounds.  Biomass is the value calculated by
* clDimensionAnalysis.
*
* Biomass stem fraction = exp(B1 + B2/DBH), where DBH is in cm and B1 and B2
* are parameters.
*
* This creates a new tree float data member called "FIA_volume".
*
* The name string and parameter file call string for this behavior are both
* "FIA Merchantable Stem Wood Volume".
*
* This behavior may only be applied to adults.
*
* NOTE: Merchantable stem wood volume was put off to probably occur in a post-hoc
* analysis tool.
*
* Copyright 2010 Charles D. Canham.
* @author Lora E. Murphy
*
* <br>Edit history:
* <br>-----------------
* <br>November 18, 2010 - Created (LEM)
*/
class clMerchantableStemWoodVolume : virtual public clBehaviorBase {
  public:

  /**
  * Constructor.
  * @param p_oSimManager clSimManager object.
  */
  clMerchantableStemWoodVolume(clSimManager *p_oSimManager);

  /**
  * Destructor.  Deletes arrays.
  */
  ~clMerchantableStemWoodVolume();

  /**
  * Makes volume calculations.  A query is sent to the tree population to get
  * all trees to which this behavior is applied.  For each, the stem wood volume
  * is calculated and the value is placed in the "FIA_volume" float tree data
  * member.
  */
  void Action();

  /**
  * Does setup for this behavior.  Steps:
  * <ol>
  * <li>Reads values from the parameter file and validates them.</li>
  * <li>Formats the query string, which is used to do tree searches.</li>
  * <li>Gets the return codes for the "Biomass" tree float data member.</li>
  * <li>Calls Action() so that the initial conditions volume will be added</li>
  * </ol>
  * @param p_oDoc DOM tree of parsed input file.
  * @throw modelErr If any tree does not have the "Biomass" float data member
  * registered
  */
  void GetData(xercesc::DOMDocument *p_oDoc);

  /**
  * Registers the "FIA_volume" float data member.  The return codes are captured
  * in the mp_iVolumeCodes array.
  * @throw modelErr if this behavior is being applied to any tree type except
  * adults.
  */
  void RegisterTreeDataMembers();

  protected:
  /** B1 in the stem fraction equation. Array size is # species to which
   * this behavior applies.*/
  float *mp_fB1;

  /** B2 in the stem fraction equation. Array size is # species to which
   * this behavior applies.*/
  float *mp_fB2;

  /** WOOD_SPGR_GREENVOL_DRYWT in the volume equation. Array size is # species
   * to which this behavior applies.*/
  float *mp_fSpgr;

  /** Helps with array access.  Array size is number of total species.*/
  int *mp_iIndexes;

  /**String to pass to clTreePopulation::Find() in order to get the trees for
  * which to calculate biomass.  This will instigate a species/type search for
  * all the species and types to which this behavior applies.*/
  char *m_cQuery;

  /**Holds data member codes for "FIA_Volume" float data member. Array
  * index is # species to which this behavior applies.*/
  short int *mp_iVolumeCodes;

  /**Holds data member codes for "Biomass" float data member. Array
  * index is # species to which this behavior applies.*/
  short int *mp_iBiomassCodes;
};
//---------------------------------------------------------------------------

#endif // MerchantableStemWoodVolume_H
